# flake8: noqa: F401
"""
pyrdp: A pure-Python RapidsDB client library.

"""

"""
0	Threads may not share the module.
1	Threads may share the module, but not connections.
2	Threads may share the module and connections.
3	Threads may share the module, connections and cursors.
"""

"""

paramstyle |	Meaning
------------------------------------------------------
qmark	   | Question mark style, e.g. ...WHERE name=?
numeric	   | Numeric, positional style, e.g. ...WHERE name=:1
named	   | Named style, e.g. ...WHERE name=:name
format	   | ANSI C printf format codes, e.g. ...WHERE name=%s
pyformat   | Python extended format codes, e.g. ...WHERE name=%(name)s
"""

from ._version import __version__
from .connections import Connection
from .constants import FIELD_TYPE
from .cursors import Cursor, SSCursor
from .err import *
from .sa.pyrdp import RDPDialect_pyrdp
from .logger import configure_logging

logger = configure_logging()

# dbapi2.0 connection object requires close(), commit(), rollback(), cursor()
# dbapi2.0 optional: autocommit()
# dbapi2.0 optional features: callproc, setinputsizes, setoutputsize

# DBAPI required attributes
threadsafety = 1
# PEP249 is version 2.0; PEP248 is version 1.0, here we follow PEP249
apilevel = "2.0"
paramstyle = "format"


class DBAPISet(frozenset):
    def __ne__(self, other):
        if isinstance(other, set):
            return frozenset.__ne__(self, other)
        else:
            return other not in self

    def __eq__(self, other):
        if isinstance(other, frozenset):
            return frozenset.__eq__(self, other)
        else:
            return other in self

    def __hash__(self):
        return frozenset.__hash__(self)


STRING = DBAPISet(
    [FIELD_TYPE.ENUM, FIELD_TYPE.STRING, FIELD_TYPE.VAR_STRING, FIELD_TYPE.VARCHAR]
)
BINARY = DBAPISet(
    [
        FIELD_TYPE.BLOB,
        FIELD_TYPE.LONG_BLOB,
        FIELD_TYPE.MEDIUM_BLOB,
        FIELD_TYPE.TINY_BLOB,
    ]
)
NUMBER = DBAPISet(
    [
        FIELD_TYPE.DECIMAL,
        FIELD_TYPE.DOUBLE,
        FIELD_TYPE.FLOAT,
        FIELD_TYPE.INT24,
        FIELD_TYPE.LONG,
        FIELD_TYPE.LONGLONG,
        FIELD_TYPE.TINY,
        FIELD_TYPE.YEAR,
    ]
)
DATE = DBAPISet([FIELD_TYPE.DATE, FIELD_TYPE.NEWDATE])
TIME = DBAPISet([FIELD_TYPE.TIME])
TIMESTAMP = DBAPISet([FIELD_TYPE.TIMESTAMP, FIELD_TYPE.DATETIME])
DATETIME = TIMESTAMP
NULL = "NULL"


def Binary(x):
    """Return x as a binary type."""
    return bytes(x)


def connect(*args, **kwargs):
    from .connections import Connection

    return Connection(*args, **kwargs)


__all__ = [
    "BINARY",
    "Binary",
    "Connection",
    "DATE",
    "DATETIME",
    "DataError",
    "DatabaseError",
    "Error",
    "FIELD_TYPE",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "RDBError",
    "NULL",
    "NUMBER",
    "NotSupportedError",
    "DBAPISet",
    "OperationalError",
    "ProgrammingError",
    "STRING",
    "TIME",
    "TIMESTAMP",
    "Warning",
    "apilevel",
    "constants",
    "converters",
    "connect",
    "paramstyle",
    "threadsafety",
    "__version__",
]

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



#
# Autogenerated by Thrift Compiler (0.20.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
#import ColumnMetadata.ttypes
from .. import ColumnMetadata

from thrift.transport import TTransport
all_structs = []


class RowMetadataResponse(object):
    """
    Attributes:
     - numColumns
     - columns
     - isBoundedResult

    """


    def __init__(self, numColumns=None, columns=None, isBoundedResult=None,):
        self.numColumns = numColumns
        self.columns = columns
        self.isBoundedResult = isBoundedResult

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.numColumns = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.columns = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = ColumnMetadata.ttypes.ColumnMetadata()
                        _elem5.read(iprot)
                        self.columns.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.BOOL:
                    self.isBoundedResult = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('RowMetadataResponse')
        if self.numColumns is not None:
            oprot.writeFieldBegin('numColumns', TType.I32, 1)
            oprot.writeI32(self.numColumns)
            oprot.writeFieldEnd()
        if self.columns is not None:
            oprot.writeFieldBegin('columns', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.columns))
            for iter6 in self.columns:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.isBoundedResult is not None:
            oprot.writeFieldBegin('isBoundedResult', TType.BOOL, 3)
            oprot.writeBool(self.isBoundedResult)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.numColumns is None:
            raise TProtocolException(message='Required field numColumns is unset!')
        if self.columns is None:
            raise TProtocolException(message='Required field columns is unset!')
        if self.isBoundedResult is None:
            raise TProtocolException(message='Required field isBoundedResult is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(RowMetadataResponse)
RowMetadataResponse.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'numColumns', None, None, ),  # 1
    (2, TType.LIST, 'columns', (TType.STRUCT, [ColumnMetadata.ttypes.ColumnMetadata, None], False), None, ),  # 2
    (3, TType.BOOL, 'isBoundedResult', None, None, ),  # 3
)
fix_spec(all_structs)
del all_structs



# flake8: noqa: E501
from typing import TYPE_CHECKING, Any, Optional, Type

from sqlalchemy import exc, types, util
from sqlalchemy.engine import default, reflection
from sqlalchemy.sql import compiler, sqltypes
from sqlalchemy.sql import text as satext
from sqlalchemy.sql.elements import quoted_name

from . import types as RDP_types

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL

str_type_map = {
    "BOOLEAN": sqltypes.BOOLEAN,
    "INTEGER": sqltypes.INTEGER,
    "DECIMAL": sqltypes.DECIMAL,
    "FLOAT": sqltypes.FLOAT,
    "DATE": sqltypes.DATE,
    "VARCHAR": sqltypes.VARCHAR,
    "VARBINARY": sqltypes.VARBINARY,
    "TIMESTAMP": sqltypes.TIMESTAMP,
}

# All keywords from parser file, and not sure if all keyword include in below.
RESERVED_WORDS = {
    "add",
    "all",
    "and",
    "as",
    "asc",
    "between",
    "both",
    "by",
    "catalog",
    "connector",
    "corresponding",
    "create",
    "cross",
    "current_catalog",
    "current_federation",
    "current_schema",
    "current_timestamp",
    "data",
    "date",
    "delete",
    "desc",
    "disabled",
    "distinct",
    "drop",
    "enabled",
    "enable",
    "escape",
    "except",
    "false",
    "federation",
    "for",
    "full",
    "from",
    "group",
    "having",
    "in",
    "index",
    "insert",
    "intersect",
    "into",
    "is",
    "join",
    "leading",
    "left",
    "like",
    "limit",
    "no",
    "node",
    "not",
    "now",
    "null",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "over",
    "overlay",
    "partition",
    "reference",
    "refresh",
    "right",
    "schema",
    "select",
    "set",
    "stats",
    "statistics",
    "table",
    "timestamp",
    "top",
    "true",
    "union",
    "update",
    "using",
    "where",
    "with",
    "boolean",
    "decimal",
    "float",
    "integer",
    "to",
    "trailing",
    "type",
    "values",
    "varbinary",
    "varchar",
    "abs",
    "array_element",
    "array_length",
    "avg",
    "case",
    "cast",
    "ceiling",
    "ceil",
    "char",
    "char_length",
    "coalesce",
    "concat",
    "count",
    "day",
    "dayofmonth",
    "dayofweek",
    "dayofyear",
    "decode",
    "else",
    "end",
    "exists",
    "exp",
    "extract",
    "field",
    "floor",
    "format_currency",
    "from_unixtime",
    "hour",
    "if",
    "ifnull",
    "interval",
    "key",
    "lower",
    "ltrim",
    "max",
    "min",
    "minute",
    "mod",
    "monotonic",
    "month",
    "nullif",
    "octet_length",
    "position",
    "pow",
    "power",
    "primary",
    "quarter",
    "rand",
    "repeat",
    "replace",
    "round",
    "row_number",
    "rtrim",
    "second",
    "set_field",
    "since_epoch",
    "space",
    "sqrt",
    "stddev",
    "stddev_pop",
    "stddev_samp",
    "substr",
    "substring",
    "sum",
    "then",
    "to_dtinterval",
    "to_timestamp",
    "to_yminterval",
    "trim",
    "truncate",
    "typeof",
    "unique",
    "upper",
    "variance",
    "var_pop",
    "var_samp",
    "week",
    "weekday",
    "weekofyear",
    "when",
    "xtypeof",
    "year",
    "hint_start",
    "hint_end",
    "block",
    "bloom",
    "btree",
    "hash",
    "loop",
    "skiplist",
    "tree",
    "value",
    "__timestamp",
}

# __timestamp is a magic word for superset

# replace sqlalchemy defined data type
colspecs = {
    types.BOOLEAN: RDP_types.BOOLEAN,
    types.INTEGER: RDP_types.INTEGER,
    types.DECIMAL: RDP_types.DECIMAL,
    types.FLOAT: RDP_types.FLOAT,
    types.TIMESTAMP: RDP_types.DATETIME,
    types.VARCHAR: RDP_types.VARCHAR,
    types.DateTime: RDP_types.DATETIME,
    types.VARBINARY: RDP_types.VARBINARY,
    types.DATETIME: RDP_types.DATETIME,
    types.DATE: RDP_types.DATE,
}

# ?
ischema_names = {
    "boolean": RDP_types.BOOLEAN,
    "integer": RDP_types.INTEGER,
    "decimal": RDP_types.DECIMAL,
    "float": RDP_types.FLOAT,
    "datetime": RDP_types.DATETIME,
    "varchar": RDP_types.VARCHAR,
    "varbinary": RDP_types.VARBINARY,
}

type_map: dict[int, Any] = {
    0: sqltypes.NULLTYPE,
    1: sqltypes.BOOLEAN,
    2: sqltypes.INTEGER,
    3: sqltypes.SMALLINT,
    4: sqltypes.INTEGER,
    5: sqltypes.INTEGER,
    6: sqltypes.DECIMAL,
    7: sqltypes.FLOAT,
    8: sqltypes.DATE,
    9: sqltypes.TIME,
    10: sqltypes.TIMESTAMP,
    11: sqltypes.Interval,
    12: sqltypes.VARCHAR,
    13: sqltypes.VARBINARY,
}


class RDPIdentifierPreparer(compiler.IdentifierPreparer):

    reserved_words = RESERVED_WORDS

    def __init__(
        self,
        dialect,
        server_ansiquotes=False,
        quote_case_sensitive_collations=False,
        **kw,
    ):

        quote = '"'

        # super(RDPIdentifierPreparer, self).__init__(
        super().__init__(
            dialect,
            initial_quote=quote,
            escape_quote=quote,
            quote_case_sensitive_collations=quote_case_sensitive_collations,
        )

    def format_table(self, table, use_schema=True, name=None):
        """Prepare a quoted table and schema name."""
        if name is None:
            name = table.name

        # 加引号可以保证 表明区分大小写
        # result = self.quote(name)
        result = '"' + name + '"'

        effective_schema = self.schema_for_object(table)

        if effective_schema is None:
            effective_schema = self.dialect.default_schema_name.upper()

        if not self.omit_schema and use_schema and effective_schema:
            result = self.quote_schema(effective_schema) + "." + result

        # RDP support catalog
        catalog = self._get_catalog()

        if catalog:
            result = self._quote_catalog(catalog) + "." + result

        return result

    # def format_column(
    #     self,
    #     column,
    #     use_table=False,
    #     name=None,
    #     table_name=None,
    #     use_schema=False,
    # ):
    #     if name is None:
    #         name = column.name
    #     if name.lower() in self.reserved_words:
    #         name = name.upper()
    #     if not getattr(column, "is_literal", False):
    #         if use_table:
    #             return (
    #                 self.format_table(
    #                     column.table, use_schema=use_schema, name=table_name
    #                 )
    #                 + "."
    #                 + self.quote(name)
    #             )
    #         else:
    #             return self.quote(name)
    #     else:
    #         if use_table:
    #             return (
    #                 self.format_table(
    #                     column.table, use_schema=use_schema, name=table_name
    #                 )
    #                 + "."
    #                 + name
    #             )
    #         else:
    #             return name

    def _quote_catalog(self, catalog):
        return catalog

    def _get_catalog(self):
        return self.dialect.default_catalog_name


class RDPStatementCompiler(compiler.SQLCompiler):
    reserved_words = RESERVED_WORDS

    def visit_table(self, table, **kwargs):
        """table is a Table object"""
        effective_schema = self.preparer.schema_for_object(table)
        table.name = quoted_name(table.name, True)
        if not effective_schema:
            effective_schema = self.dialect.default_schema_name
        ret = (
            self.preparer.quote_schema(effective_schema)
            + "."
            + self.preparer.quote_identifier(table.name)
        )
        return ret

    def visit_column(self, column, add_to_result_map=None, include_table=True, **kwargs):
        name = orig_name = column.name
        if name is None:
            name = self._fallback_column_name(column)

        # TODO  visit_column and format_column need handle this
        if name.lower() in self.reserved_words:
            name = name.upper()

        is_literal = column.is_literal
        # if not is_literal and isinstance(name, elements._truncated_label):
        #     name = self._truncated_identifier("colident", name)

        if add_to_result_map is not None:
            add_to_result_map(name, orig_name, (column, name, column.key), column.type)

        if is_literal:
            name = self.escape_literal_column(name)
        else:
            name = self.preparer.quote(name)

        if column.table is not None:
            table_name = column.table.name
            name = self.preparer.quote_identifier(table_name) + "." + name
        return name

    def limit_clause(self, select, **kwargs):
        text = ""
        if select._limit_clause is not None:
            text += "\n LIMIT " + self.process(select._limit_clause, **kwargs)
        if select._offset_clause is not None:
            if select._limit_clause is None:
                # 2147384648 is the max. no. of records per result set
                text += "\n LIMIT 2147384648"
            text += " OFFSET " + self.process(select._offset_clause, **kwargs)
        return text


class RDPTypeCompiler(compiler.GenericTypeCompiler):

    def visit_BOOLEAN(self, type_, **kw):
        # type_.create_constraint = False
        return "BOOLEAN"

    def visit_TINYINT(self, type_, **kw):
        # if type_.display_width is not None:
        #     return "TINYINT(%s)" % type_.display_width
        # else:
        return "TINYINT"

    def visit_SMALLINT(self, type_, **kw):
        # if type_.display_width is not None:
        #     return "SMALLINT(%(display_width)s)" % {'display_width': type_.display_width}
        # else:
        return "SMALLINT"

    def visit_INTEGER(self, type_, **kw):
        # if type_.display_width is not None:
        #     return  "INTEGER(%(display_width)s)" % {'display_width': type_.display_width}
        # else:
        return "INTEGER"

    def visit_BIGINT(self, type_, **kw):
        # if type_.display_width is not None:
        #     return "BIGINT(%(display_width)s)" % {'display_width': type_.display_width}
        # else:
        return "INTEGER"

    def visit_DECIMAL(self, type_, **kw):
        if type_.precision is None:
            return "DECIMAL"
        elif type_.scale is None:
            return "DECIMAL(%(precision)s)" % {"precision": type_.precision}
        else:
            return "DECIMAL(%(precision)s, %(scale)s)" % {
                "precision": type_.precision,
                "scale": type_.scale,
            }

    def visit_FLOAT(self, type_, **kw):
        if type_.scale is not None and type_.precision is not None:
            return "FLOAT(%s, %s)" % (type_.precision, type_.scale)
        elif type_.precision is not None:
            return "FLOAT(%s)" % (type_.precision,)
        else:
            return "FLOAT"

    def visit_NUMERIC(self, type_, **kw):
        return self.visit_DECIMAL(type_, **kw)

    def visit_DATE(self, type_, **kw):
        return "DATE"

    def visit_TIME(self, type_, **kw):
        if getattr(type_, "fsp", None):
            return "TIME(%d)" % type_.fsp
        else:
            return "TIME"

    def visit_TIMESTAMP(self, type_, **kw):
        # print("robin-timestamp")
        if getattr(type_, "fsp", None):
            return "TIMESTAMP(%d)" % type_.fsp
        else:
            return "TIMESTAMP"

    def visit_VARCHAR(self, type_, **kw):
        if type_.length:
            return "VARCHAR(%d)" % type_.length
        else:
            return "VARCHAR"

    def visit_VARBINARY(self, type_, **kw):
        return "VARBINARY(%d)" % type_.length

    def visit_CHAR(self, type_, **kw):
        return "CHAR"

    # RDP do not support DateTime, this is just past some test case
    def visit_DATETIME(self, type_, **kw):
        return "DATE"

    def visit_TEXT(self, type_, **kw):
        return "VARCHAR"


class RDPDDLCompiler(compiler.DDLCompiler):

    def visit_create_table(self, create):
        table = create.element
        preparer = self.preparer

        text = "\nCREATE "
        if table._prefixes:
            text += " ".join(table._prefixes) + " "
        text += "TABLE " + preparer.format_table(table) + " "

        create_table_suffix = self.create_table_suffix(table)
        if create_table_suffix:
            text += create_table_suffix + " "

        text += "("

        separator = "\n"

        # if only one primary key, specify it along with the column
        first_pk = False
        for create_column in create.columns:
            column = create_column.element
            try:
                processed = self.process(
                    create_column, first_pk=column.primary_key and not first_pk
                )
                if processed is not None:
                    text += separator
                    separator = ", \n"
                    text += "\t" + processed
                if column.primary_key:
                    first_pk = True
            except exc.CompileError as ce:
                util.raise_from_cause(
                    exc.CompileError(
                        util.u("(in table '%s', column '%s'): %s")
                        % (table.description, column.name, ce.args[0])
                    )
                )

        const = self.create_table_constraints(
            table,
            _include_foreign_key_constraints=create.include_foreign_key_constraints,  # noqa
        )
        if const:
            text += separator + "\t" + const

        text += "\n)%s" % self.post_create_table(table)
        return text

    # RDP do not support check constraint for boolean, and just return None here
    def visit_check_constraint(self, constraint):
        return None


class RDPExecutionContext(default.DefaultExecutionContext):
    pass


class RDPDialect(default.DefaultDialect):
    name = "rapidsdb"
    default_paramstyle = "format"

    statement_compiler = RDPStatementCompiler
    type_compiler = RDPTypeCompiler
    ddl_compiler = RDPDDLCompiler
    preparer: Type[RDPIdentifierPreparer] = RDPIdentifierPreparer
    execution_ctx_cls = RDPExecutionContext

    encoding = "utf-8"
    supports_statement_cache = True
    supports_native_boolean = True
    supports_alter = False
    supports_views = False
    convert_unicode = True
    supports_unicode_statements = True
    supports_unicode_binds = True
    requires_name_normalize = True

    postfetch_lastrowid = False
    supports_empty_insert = False
    supports_sane_rowcount = False
    supports_sane_multi_rowcount = False
    isolation_level = "REPEATABLE READ"
    supports_default_values = False
    supports_savepoints = False

    # A dictionary of TypeEngine classes from sqlalchemy.types mapped to subclasses that are specific to the dialect class
    colspecs = colspecs
    ischema_names = ischema_names
    type_map = type_map

    def __init__(self, isolation_level=None, auto_convert_lobs=True, **kwargs):
        super().__init__(**kwargs)
        self.isolation_level = isolation_level
        self.auto_convert_lobs = auto_convert_lobs
        self.identifier_preparer = self.preparer(self)
        self.default_schema_name = "SYSTEM"
        self.default_catalog_name = "RAPIDS"

    def create_connect_args(self, url: "URL") -> tuple:
        # create a dict from url options and change key 'username' to 'user'
        db_opts: dict = url.translate_connect_args(username="user")
        db_opts["user"] = db_opts["user"].upper()
        db_opts.setdefault("port", 4333)
        db_opts.update(url.query)
        meta_dict = self._parse_url(db_opts.setdefault("database", "RAPIDS/SYSTEM"))
        table_metadata = TableMetaData(meta_dict["catalog"], meta_dict["schema"])

        # kwargs.setdefault("federation", meta_dict["federation"])
        db_opts.setdefault("catalog", meta_dict["catalog"])
        db_opts.setdefault("schema", meta_dict["schema"])
        db_opts.setdefault("database", meta_dict["schema"])

        if table_metadata.get_catalog():
            self._set_default_catalog(table_metadata.get_catalog())
        if table_metadata.get_schema():
            self._set_default_schema(table_metadata.get_schema())

        return (), db_opts

    def _parse_url(self, db_url: str) -> dict:
        """Generate a dict of federation, catalog, schema names from db_url."""
        meta_dict: dict[str, Optional[str]] = {
            "federation": None,
            "catalog": None,
            "schema": None,
        }

        names = db_url.split("/")
        nums = len(names)
        if nums == 3:
            meta_dict["federation"] = names[0]
            meta_dict["catalog"] = names[1]
            meta_dict["schema"] = names[2]
        elif nums == 2:
            meta_dict["catalog"] = names[0]
            meta_dict["schema"] = names[1]
        elif nums == 1:
            meta_dict["catalog"] = meta_dict["schema"] = names[0]
        else:
            pass

        return meta_dict

    def on_connect(self):
        super_ = super().on_connect()

        def on_connect(conn):
            if super_ is not None:
                super_(conn)
            cursor = conn.cursor()
            cursor.close()

        return on_connect

    # _isolation_lookup = {
    #     "SERIALIZABLE",
    #     "READ UNCOMMITTED",
    #     "READ COMMITTED",
    #     "REPEATABLE READ",
    # }

    # does not work with pyrdpdb currently
    # def set_isolation_level(self, connection, level):
    #     if level == "AUTOCOMMIT":
    #         pass
    #     else:
    #         # no need to set autocommit false explicitly,since it is false by default
    #         if level not in self._isolation_lookup:
    #             raise exc.ArgumentError(
    #                 "Invalid value '%s' for isolation_level. "
    #                 "Valid isolation levels for %s are %s"
    #                 % (level, self.name, ", ".join(self._isolation_lookup))
    #             )
    #         else:
    #             pass

    # def get_isolation_level(self, connection):
    #     return self.isolation_level

    # Important ,but stil not know why TODO
    def _check_unicode_returns(self, connection):
        return True

    def _check_unicode_description(self, connection):
        return True

    # RDP need set a default schema for read and write data
    def _get_default_schema_name(self, connection):
        return self.default_schema_name

    def _set_default_schema(self, schema_name):
        self.default_schema_name = schema_name

    def _set_default_catalog(self, catalog_name):
        self.default_catalog_name = catalog_name

    def normalize_name(self, name):
        if name is None:
            return None

        if name.upper() == name and not self.identifier_preparer._requires_quotes(
            name.lower()
        ):
            name = name.lower()
        elif name.lower() == name:
            return quoted_name(name, quote=True)

        return name

    def denormalize_name(self, name):
        if name is None:
            return None

        if name.lower() == name and not self.identifier_preparer._requires_quotes(
            name.lower()
        ):
            name = name.upper()
        return name

    def has_table(self, connection, table_name, schema=None, **kw):
        """Return whether a specified table exists.

        Note: it is a synchronous function, requires sync connection.
        """
        # schema = schema or self.default_schema_name

        # how to make sure a table if exist in RDP ?
        qname = self._quote_string(table_name)
        # upper_table_name = qname.upper()
        if schema is not None:
            upper_schema = self._quote_string(schema.upper())

            result = connection.execute(
                satext(
                    "SELECT 1 FROM RAPIDS.SYSTEM.TABLES WHERE SCHEMA_NAME=%s AND TABLE_NAME=%s;"
                    % (upper_schema, qname)
                )
            )
        else:
            result = connection.execute(
                satext(
                    "SELECT 1 FROM RAPIDS.SYSTEM.TABLES WHERE TABLE_NAME=%s;" % (qname)
                )
            )

        table_exists = bool(result.first())
        return table_exists

    def has_sequence(self, connection, sequence_name, schema=None):
        """RDP do not support sequence"""
        return False

    @reflection.cache
    def get_schema_names(self, connection, **kwargs):
        """get all schemas of the db system"""
        result = connection.execute(
            satext("SELECT SCHEMA_NAME FROM RAPIDS.SYSTEM.SCHEMAS;")
        )

        return list([name for name, in result.fetchall()])

    @reflection.cache
    def get_table_names(self, connection, schema=None, **kwargs):
        """get all columns info of a specified schema"""
        schema = schema or self.default_schema_name
        schema = "'" + schema.upper() + "'"
        result = connection.execute(
            satext(
                f"SELECT TABLE_NAME FROM RAPIDS.SYSTEM.TABLES WHERE SCHEMA_NAME={schema};"
            )
        )

        tables = list(
            [
                # self.normalize_name(row[0]) for row in result.fetchall()
                row[0]
                for row in result.fetchall()
            ]
        )
        return tables

    def get_temp_table_names(self, connection, schema=None, **kwargs):
        """RDP do not support temp table"""
        return []

    def get_view_names(self, connection, schema=None, **kwargs):
        """RDP do not support view"""
        return []

    def get_view_definition(self, connection, view_name, schema=None, **kwargs):
        """RDP do not support view"""
        return ""

    @staticmethod
    def resolve_tablename(table_name) -> str:
        if table_name == table_name.lower():  # this is lowercase tablename
            query_table_name = table_name.upper()
        else:
            query_table_name = table_name
        return query_table_name

    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kwargs):
        """get all columns info of a specified table"""

        schema = schema or self.default_schema_name
        table_name = self._quote_string(table_name)
        query_table_name = self.resolve_tablename(table_name)
        if schema is not None:
            schema = self._quote_string(schema).upper()
            result = connection.execute(
                satext(
                    "SELECT COLUMN_NAME, DATA_TYPE, IS_NULLABLE, SCALE "
                    "FROM RAPIDS.SYSTEM.COLUMNS "
                    "WHERE SCHEMA_NAME=%s AND TABLE_NAME=%s ORDER BY ORDINAL;"
                    % (schema, query_table_name)
                )
            )
        else:
            result = connection.execute(
                satext(
                    "SELECT COLUMN_NAME, DATA_TYPE, IS_NULLABLE, SCALE "
                    "FROM RAPIDS.SYSTEM.COLUMNS WHERE TABLE_NAME=%s ORDER BY ORDINAL;"
                    % (query_table_name)
                )
            )

        columns = []
        for row in result.fetchall():
            column = {"name": row[0], "nullable": row[2] == "TRUE"}
            # magic
            if hasattr(types, row[1]):
                if row[1] == "TIMESTAMP":
                    column["type"] = getattr(RDP_types, "DATETIME")
                else:
                    column["type"] = getattr(RDP_types, row[1])
            elif hasattr(RDP_types, row[1]):
                if row[1] == "TIMESTAMP":
                    column["type"] = getattr(types, "DATETIME")
                else:
                    column["type"] = getattr(types, row[1])
            else:
                util.warn(
                    "Did not recognize type '%s' of column '%s'"
                    % (row[1], column["name"])
                )
                column["type"] = types.NULLTYPE

            columns.append(column)
        # print(columns)
        return columns

    def get_multi_columns(self, connection, schema, filter_names, **kwargs):
        """Override default get_multi_columns to integrate with RapidsDB metadata retrival."""
        _schema = schema or self.default_schema_name
        result = {}

        table_filter = ""
        params = {"schema": _schema.upper()}
        metadata_query = (
            "SELECT SCHEMA_NAME, TABLE_NAME, COLUMN_NAME, DATA_TYPE, IS_NULLABLE, PRECISION, SCALE, COMMENT "
            "FROM RAPIDS.SYSTEM.COLUMNS "
            "WHERE SCHEMA_NAME = :schema {table_filter}"
        )
        if filter_names:
            table_filter = "AND TABLE_NAME IN :filter_names"
            query_filter_names = filter_names
            # query_filter_names = [self.resolve_tablename(n) for n in filter_names]
            params["filter_names"] = tuple(query_filter_names)
        stmt = metadata_query.format(table_filter=table_filter)
        rset = connection.execute(satext(stmt), params).fetchall()

        for row in rset:
            result_schema_name = row[0]  # for "SCHEMA_NAME"
            result_table_name = row[1]  # "TABLE_NAME"
            coltype = str_type_map.setdefault(row[3], sqltypes.NullType)  # 3: "DATA_TYPE"
            precision = int(row[5]) if row[5] else None
            scale = int(row[6]) if row[6] else None
            if coltype == sqltypes.NUMERIC:
                coltype = sqltypes.NUMERIC(precision, scale)
            elif coltype == sqltypes.DECIMAL:
                coltype = sqltypes.DECIMAL(precision, scale)
            cdict = {
                "name": row[2].lower(),
                "type": coltype,
                "nullable": bool(row[4]),
                "comment": row[7],
            }  # 2: "COLUMN_NAME", 4: "IS_NULLABLE"

            orig_tname = result_table_name
            key = (schema, result_table_name) if schema else (None, orig_tname)
            if key not in result:
                result[key] = []
            result[key].append(cdict)
        return result.items()

    def get_foreign_keys(self, connection, table_name, schema=None, **kwargs):
        """RDP do not support foreign keys"""
        return []

    def get_indexes(self, connection, table_name, schema=None, **kwargs):
        """get all indexes info of a specified table"""
        schema = schema or self.default_schema_name
        table_name = "'" + table_name + "'"

        if schema is not None:
            schema = "'" + schema + "'"
            result = connection.execute(
                satext(
                    "SELECT INDEX_NAME, IS_UNIQUE, COLUMN_NAME FROM  RAPIDS.SYSTEM.INDEXES WHERE SCHEMA_NAME=%s AND TABLE_NAME=%s ORDER BY ORDINAL;"
                    % (schema, table_name)
                )
            )
        else:
            result = connection.execute(
                satext(
                    "SELECT INDEX_NAME, IS_UNIQUE, COLUMN_NAME FROM  RAPIDS.SYSTEM.INDEXES WHERE TABLE_NAME=%s ORDER BY ORDINAL;"
                    % (table_name,)
                )
            )

        indexes = {}

        if result is not None:
            try:
                for row in result.fetchall():
                    name = row[0]
                    unique = row[1]
                    column = row[2]

                    name = self.normalize_name(name)
                    column = self.normalize_name(column)

                    if name not in indexes:
                        indexes[name] = {
                            "name": name,
                            "unique": unique,
                            "column_names": [column],
                        }
                    else:
                        indexes[name]["column_names"].append(column)
            except BaseException:
                print("Get indexe error.")

        return list(indexes.values())

    def get_pk_constraint(self, connection, table_name, schema=None, **kwargs):
        """RDP do not support constraint"""
        return []

    def get_unique_constraints(self, connection, table_name, schema=None, **kwargs):
        """RDP do not support unique constraints"""
        return []

    def get_check_constraints(self, connection, table_name, schema=None, **kwargs):
        """RDP do not support check constraints"""
        return []

    def get_table_oid(self, connection, table_name, schema=None, **kwargs):
        pass

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        """RDP do not support table comment"""
        return {}

    def _quote_string(self, string) -> str:

        return "'" + string + "'"

    def do_savepoint(self, connection, name):
        """Require "SAVEPOINT" support from DB"""
        raise NotImplementedError("SAVEPOINT is not supported yet")

    def do_release_savepoint(self, connection, name):
        """Require "SAVEPOINT" support from DB"""
        raise NotImplementedError("SAVEPOINT is not supported yet")

    def do_rollback_to_savepoint(self, connection, name):
        """Require "SAVEPOINT" support from DB"""
        raise NotImplementedError("SAVEPOINT is not supported yet")


class TableMetaData(object):
    def __init__(self, catalog=None, schema=None, name=None):
        self.catalog = catalog
        self.schema = schema
        self.name = name

    def get_catalog(self):
        return self.catalog

    def get_schema(self):
        return self.schema

    def get_name(self):
        return self.name



# -*- coding: utf-8 -*-
import struct
import time
from functools import wraps
from collections.abc import Coroutine

import logging

logger = logging.getLogger("pyrdp.utils")


def timeit(func):
    """Decorator to time function execution time."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        results = func(*args, **kwargs)
        end_time = time.time()
        runtime = round((end_time - start_time), 4)
        print(f"{runtime:.4f} seconds.")

        return results

    return wrapper


def int2_4bytes(i):
    """
    Encode int to 4 bytes.
    :param number i : input int .
    :return : 4 bytes
    """
    return struct.pack(">I", i)


def bytes2int(bs):
    """
    Decode bytes to int.
    :param bytes bs : input bytes .
    :return : int
    """
    return struct.unpack(">I", bs)


def int2byte(i):
    """
    Encode int to byte.
    :param number i : input int .
    :return : byte
    """
    return struct.pack("!B", i)


def byte2int(b):
    """
    Decode byte to int.
    :param byte b : input byte .
    :return : int
    """
    if isinstance(b, int):
        return b
    else:
        return struct.unpack("!B", b)[0]


def join_bytes(bs):
    if len(bs) == 0:
        return ""
    else:
        rv = bs[0]
        for b in bs[1:]:
            rv += b
        return rv


def _pack_int24(n):
    return struct.pack("<I", n)[:3]


def _lenenc_int(i):
    if i < 0:
        raise ValueError(
            "Encoding %d is less than 0 - no representation in LengthEncodedInteger" % i
        )
    elif i < 0xFB:
        return bytes([i])
    elif i < (1 << 16):
        return b"\xfc" + struct.pack("<H", i)
    elif i < (1 << 24):
        return b"\xfd" + struct.pack("<I", i)[:3]
    elif i < (1 << 64):
        return b"\xfe" + struct.pack("<Q", i)
    else:
        raise ValueError(
            "Encoding %x is larger than %x - no representation in LengthEncodedInteger"
            % (i, (1 << 64))
        )


class _ContextManager(Coroutine):

    __slots__ = ("_coro", "_obj")

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    @property
    def gi_frame(self):
        return self._coro.gi_frame

    @property
    def gi_running(self):
        return self._coro.gi_running

    @property
    def gi_code(self):
        return self._coro.gi_code

    def __next__(self):
        return self.send(None)

    def __iter__(self):
        return self._coro.__await__()

    def __await__(self):
        return self._coro.__await__()

    async def __aenter__(self):
        self._obj = await self._coro
        return self._obj

    async def __aexit__(self, exc_type, exc, tb):
        await self._obj.close()
        self._obj = None


class _ConnectionContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc, tb):
        if exc_type is not None:
            self._obj.close()
        else:
            await self._obj.ensure_closed()
        self._obj = None


class _PoolContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc, tb):
        self._obj.close()
        await self._obj.wait_closed()
        self._obj = None


class _SAConnectionContextManager(_ContextManager):
    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._obj is None:
            self._obj = await self._coro

        try:
            return await self._obj.__anext__()
        except StopAsyncIteration:
            await self._obj.close()
            self._obj = None
            raise


class _TransactionContextManager(_ContextManager):
    async def __aexit__(self, exc_type, exc, tb):
        if exc_type:
            # await self._obj.rollback()
            pass
        else:
            if self._obj.is_active:
                # await self._obj.commit()
                pass
        self._obj = None


class _PoolAcquireContextManager(_ContextManager):

    __slots__ = ("_coro", "_conn", "_pool")

    def __init__(self, coro, pool):
        self._coro = coro
        self._conn = None
        self._pool = pool

    async def __aenter__(self):
        logger.info("Entering context, acquiring connection...")
        self._conn = await self._coro
        logger.debug("Connection acquired:", self._conn)
        return self._conn

    async def __aexit__(self, exc_type, exc, tb):
        try:
            await self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None


class _PoolConnectionContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    connection around a block:

        with (yield from pool) as conn:
            cur = yield from conn.cursor()

    while failing loudly when accidentally using:

        with pool:
            <block>
    """

    __slots__ = ("_pool", "_conn")

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        assert self._conn
        return self._conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None

    async def __aenter__(self):
        assert not self._conn
        self._conn = await self._pool.acquire()
        return self._conn

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None



#!/usr/bin/env python
import sys
from typing import Optional

from sqlalchemy import MetaData, Table, create_engine, select
from sqlalchemy.orm import Session


def main(table: str, host: str) -> None:
    engine = create_engine(f"rapidsdb+pyrdp://RAPIDS:rapids@{host}:4333/sf1", echo=False)
    meta = MetaData()
    session = Session(engine)
    t1: Optional[Table]
    if "." in table:
        schema, table = table.split(".")
        meta.reflect(bind=engine, schema=schema)
        t1 = meta.tables.get(table)
    else:
        table = table
        meta.reflect(bind=engine)
        t1 = meta.tables.get(table)
    if t1 is None:
        print("table not found")
        sys.exit(1)

    print("ready to run query")
    for r in session.execute(select(t1)):
        print(r)


if __name__ == "__main__":
    if len(sys.argv) >= 3:
        table = sys.argv[1]
        host = sys.argv[2]
    elif len(sys.argv) == 2:
        host = "localhost"
        table = sys.argv[1]
    else:
        print("Usage: python -m pyrdp.example.query <table> <host>", file=sys.stderr)
        sys.exit(-1)

    print("Connect to", host)
    main(table=table, host=host)



# -*- coding: utf-8 -*-
import logging

"""
logger component, log some userful information for debug

"""


def configure_logging(log_level=logging.ERROR, log_file_name="pyrdp.log"):
    """
    get logger
    :param appName: application name
    :param logFileName: log fiel name
    :return: logger object
    """

    logger = logging.getLogger("pyrdp")
    logger.setLevel(log_level)

    formatter = logging.Formatter("[%(levelname)s: %(name)s] %(message)s")
    # formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s") # noqa

    file_handler = logging.FileHandler(log_file_name)
    file_handler.setFormatter(formatter)
    file_handler.setLevel(log_level)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(log_level)

    file_handler.setFormatter(formatter)
    stream_handler.setFormatter(formatter)

    if not logger.hasHandlers():
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)

    logging.getLogger().setLevel(logging.WARNING)
    logger.propagate = False

    return logger



import sys

from pyrdpdb import pyrdp


def simple(host: str):
    print("\nBasic query with cursor:\n")
    conn = pyrdp.connect(host=host, port=4333, user="RAPIDS", password="rapids")
    sql = "select * from region"
    cur = conn.cursor()
    cur.execute(sql)
    res = cur.fetchall()
    print(res)


def sscursor(host: str):
    print("\nUsing SSCursor:\n")
    conn = pyrdp.connect(
        host=host,
        port=4333,
        user="RAPIDS",
        password="rapids",
        cursorclass=pyrdp.SSCursor,
    )
    with conn.cursor() as cursor:
        cursor.execute("select * from nation limit 5")
        for r in cursor:
            print(r)


def main(host: str):
    simple(host)
    sscursor(host)


if __name__ == "__main__":
    host = sys.argv[1] if len(sys.argv) > 1 else "localhost"
    main(host)



# ported from:
# https://github.com/aio-libs/aiopg/blob/master/aiopg/sa/connection.py
import weakref

from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.ddl import DDLElement
from sqlalchemy.sql.dml import UpdateBase

from ..utils import _SAConnectionContextManager, _TransactionContextManager
from . import exc
from .result import create_result_proxy


def noop(k):
    return k


# class SAConnection(Connection):
class SAConnection:

    def __init__(self, connection, engine, compiled_cache=None):
        self._connection = connection
        # self._transaction = None
        # self._savepoint_seq = 0
        self._weak_results = weakref.WeakSet()
        self._engine = engine
        self.dialect = engine.dialect
        self._compiled_cache = compiled_cache

    def execute(self, query, quiet: bool = False, *multiparams, **params):
        """Executes a SQL query with optional parameters.

        query - a SQL query string or any sqlalchemy expression.

        *multiparams/**params - represent bound parameter values to be
        used in the execution.  Typically, the format is a dictionary
        passed to *multiparams:

            await conn.execute(
                table.insert(),
                {"id":1, "value":"v1"},
            )

        ...or individual key/values interpreted by **params::

            await conn.execute(
                table.insert(), id=1, value="v1"
            )

        In the case that a plain SQL string is passed, a tuple or
        individual values in *multiparams may be passed::

            await conn.execute(
                "INSERT INTO table (id, value) VALUES (%d, %s)",
                (1, "v1")
            )

            await conn.execute(
                "INSERT INTO table (id, value) VALUES (%s, %s)",
                1, "v1"
            )

        Returns ResultProxy instance with results of SQL query
        execution.

        """
        coro = self._execute(query, quiet, *multiparams, **params)
        return _SAConnectionContextManager(coro)

    def _base_params(self, query, dp, compiled, is_update):
        """
        handle params
        """
        if dp and isinstance(dp, (list, tuple)):
            if is_update:
                dp = {c.key: pval for c, pval in zip(query.table.c, dp)}
            else:
                raise exc.ArgumentError(
                    "Don't mix sqlalchemy SELECT " "clause with positional " "parameters"
                )
        compiled_params = compiled.construct_params(dp)
        processors = compiled._bind_processors
        params = [
            {
                key: processors.get(key, noop)(compiled_params[key])
                for key in compiled_params
            }
        ]
        post_processed_params = self.dialect.execute_sequence_format(params)
        return post_processed_params[0]

    async def _executemany(self, query, dps, cursor, quiet):
        """
        executemany
        """
        result_map = None
        if isinstance(query, str):
            await cursor.executemany(query, dps, quiet)
        elif isinstance(query, DDLElement):
            raise exc.ArgumentError(
                "Don't mix sqlalchemy DDL clause and execution with parameters"
            )
        elif isinstance(query, ClauseElement):
            compiled = query.compile(dialect=self.dialect)
            params = []
            is_update = isinstance(query, UpdateBase)
            for dp in dps:
                params.append(
                    self._base_params(
                        query,
                        dp,
                        compiled,
                        is_update,
                    )
                )
            await cursor.executemany(str(compiled), params)
            result_map = compiled._result_columns
        else:
            raise exc.ArgumentError(
                "sql statement should be str or "
                "SQLAlchemy data "
                "selection/modification clause"
            )
        ret = await create_result_proxy(self, cursor, self.dialect, result_map)
        self._weak_results.add(ret)
        return ret

    async def _execute(self, query, quiet: bool = False, *multiparams, **params):
        cursor = await self._connection.cursor()
        dp = _distill_params(multiparams, params)
        if len(dp) > 1:
            return await self._executemany(query, dp, cursor, quiet=quiet)
        elif dp:
            dp = dp[0]

        result_map = None
        if isinstance(query, str):
            await cursor.execute(query, dp or None, quiet=quiet)
        elif isinstance(query, ClauseElement):
            if self._compiled_cache is not None:
                key = query
                compiled = self._compiled_cache.get(key)
                if not compiled:
                    compiled = query.compile(dialect=self.dialect)
                    if (
                        dp
                        and dp.keys() == compiled.params.keys()
                        or not (dp or compiled.params)
                    ):
                        # we only want queries with bound params in cache
                        self._compiled_cache[key] = compiled
            else:
                compiled = query.compile(
                    dialect=self.dialect(),
                    compile_kwargs={"literal_binds": True},
                )

            if not isinstance(query, DDLElement):
                # post_processed_params = self._base_params(
                #     query, dp, compiled, isinstance(query, UpdateBase)
                # )
                result_map = compiled._result_columns
            else:
                if dp:
                    raise exc.ArgumentError(
                        "Don't mix sqlalchemy DDL clause and execution with parameters"
                    )
                # post_processed_params = compiled.construct_params()
                result_map = None

            # hard remove catalog 'RAPIDS.' from entity name in query
            fixed_sql = str(compiled).replace("RAPIDS.", "")
            await cursor.execute(fixed_sql, quiet=quiet)
            # await cursor.execute(str(compiled), post_processed_params)
        else:
            raise exc.ArgumentError(
                "sql statement should be str or "
                "SQLAlchemy data "
                "selection/modification clause"
            )
        ret = await create_result_proxy(self, cursor, self.dialect, result_map)
        self._weak_results.add(ret)
        return ret

    async def scalar(self, query, *multiparams, **params):
        """Executes a SQL query and returns a scalar value."""
        res = await self.execute(query, *multiparams, **params)
        return await res.scalar()

    @property
    def closed(self):
        """The readonly property that returns True if connections is closed."""
        return self._connection is None or self._connection.closed

    @property
    def connection(self):
        return self._connection

    async def begin(self):
        coro = self._begin()
        return _TransactionContextManager(coro)

    async def _begin(self):
        if self._transaction is None:
            self._transaction = RootTransaction(self)
            return self._transaction
        else:
            return Transaction(self, self._transaction)

    async def close(self):
        """Close this SAConnection.

        This results in a release of the underlying database
        resources, that is, the underlying connection referenced
        internally. The underlying connection is typically restored
        back to the connection-holding Pool referenced by the Engine
        that produced this SAConnection. Any transactional state
        present on the underlying connection is also unconditionally
        released via calling Transaction.rollback() method.

        After .close() is called, the SAConnection is permanently in a
        closed state, and will allow no further operations.
        """
        if self._connection is None:
            return

        # if self._transaction is not None:
        #     await self._transaction.rollback()
        #     self._transaction = None
        # don't close underlying connection, it can be reused by pool
        # conn.close()
        self._engine.release(self)
        self._connection = None
        self._engine = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class Transaction:

    def __init__(self, connection, parent):
        self._connection = connection
        self._parent = parent or self
        self._is_active = True

    @property
    def is_active(self):
        """Return ``True`` if a transaction is active."""
        return self._is_active

    @property
    def connection(self):
        """Return transaction's connection (SAConnection instance)."""
        return self._connection

    async def close(self):
        """Close this transaction.

        If this transaction is the base transaction in a begin/commit
        nesting, the transaction will rollback().  Otherwise, the
        method returns.

        This is used to cancel a Transaction without affecting the scope of
        an enclosing transaction.
        """
        if not self._parent._is_active:
            return
        if self._parent is self:
            await self.rollback()
        else:
            self._is_active = False

    # async def rollback(self):
    #     """Roll back this transaction."""
    #     if not self._parent._is_active:
    #         return
    #     await self._do_rollback()
    #     self._is_active = False

    # async def _do_rollback(self):
    #     await self._parent.rollback()

    # async def commit(self):
    #     """Commit this transaction."""

    #     if not self._parent._is_active:
    #         raise exc.InvalidRequestError("This transaction is inactive")
    #     await self._do_commit()
    #     self._is_active = False

    # async def _do_commit(self):
    #     pass

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            await self.rollback()
        else:
            if self._is_active:
                await self.commit()


class RootTransaction(Transaction):

    def __init__(self, connection):
        super().__init__(connection, None)

    # async def _do_rollback(self):
    #     await self._connection._rollback_impl()

    # async def _do_commit(self):
    #     await self._connection._commit_impl()


def _distill_params(multiparams, params):
    """Given arguments from the calling form *multiparams, **params,
    return a list of bind parameter structures, usually a list of
    dictionaries.

    In the case of 'raw' execution which accepts positional parameters,
    it may be a list of tuples or lists.

    """

    if not multiparams:
        if params:
            return [params]
        else:
            return []
    elif len(multiparams) == 1:
        zero = multiparams[0]
        if isinstance(zero, (list, tuple)):
            if not zero or hasattr(zero[0], "__iter__") and not hasattr(zero[0], "strip"):
                # execute(stmt, [{}, {}, {}, ...])
                # execute(stmt, [(), (), (), ...])
                return zero
            else:
                # execute(stmt, ("value", "value"))
                return [zero]
        elif hasattr(zero, "keys"):
            # execute(stmt, {"key":"value"})
            return [zero]
        else:
            # execute(stmt, "value")
            return [[zero]]
    else:
        if hasattr(multiparams[0], "__iter__") and not hasattr(multiparams[0], "strip"):
            return multiparams
        else:
            return [multiparams]



#
# Autogenerated by Thrift Compiler (0.20.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
from .. import ColumnValue

from thrift.transport import TTransport
all_structs = []


class Row(object):
    """
    Attributes:
     - columns

    """


    def __init__(self, columns=None,):
        self.columns = columns

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.columns = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = ColumnValue.ttypes.ColumnValue()
                        _elem5.read(iprot)
                        self.columns.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Row')
        if self.columns is not None:
            oprot.writeFieldBegin('columns', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.columns))
            for iter6 in self.columns:
                iter6.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.columns is None:
            raise TProtocolException(message='Required field columns is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class RowDataResponse(object):
    """
    Attributes:
     - rows

    """


    def __init__(self, rows=None,):
        self.rows = rows

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.rows = []
                    (_etype10, _size7) = iprot.readListBegin()
                    for _i11 in range(_size7):
                        _elem12 = Row()
                        _elem12.read(iprot)
                        self.rows.append(_elem12)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('RowDataResponse')
        if self.rows is not None:
            oprot.writeFieldBegin('rows', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.rows))
            for iter13 in self.rows:
                iter13.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.rows is None:
            raise TProtocolException(message='Required field rows is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Row)
Row.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'columns', (TType.STRUCT, [ColumnValue.ttypes.ColumnValue, None], False), None, ),  # 1
)
all_structs.append(RowDataResponse)
RowDataResponse.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'rows', (TType.STRUCT, [Row, None], False), None, ),  # 1
)
fix_spec(all_structs)
del all_structs



# flake8: noqa: E501
"""
.. dialect:: pyrdp
    :name: rapidsdb
    :dbapi: pyrdp
    :connectstring: rapidsdb+pyrdp://<user>:<password>@<host>[:<port>]/<federation>/<catalog>/<schema>
"""
from sqlalchemy.dialects import registry

from .base import RDPDialect


class RDPDialect_pyrdp(RDPDialect):

    driver = "pyrdp"
    default_paramstyle = "format"
    description_encoding = None
    supports_native_decimal = True
    is_async = False
    supports_statement_cache = True
    supports_savepoints = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def import_dbapi(cls):
        from pyrdpdb import pyrdp

        pyrdp.paramstyle = cls.default_paramstyle
        return pyrdp

    def is_disconnect(self, error, connection, cursor):
        """Tells whether connection is closed.

        Parameters:
        - error: not used
        - connection: Optional[Connection]
        - cursor: not use

        Return: bool
        """
        if connection is None:
            return True
        return connection._closed


dialect = RDPDialect_pyrdp

registry.register("rapidsdb.pyrdp", "pyrdpdb.pyrdp.sa.pyrdp", "RDPDialect_pyrdp")



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class QueryExecuteResponse(object):
    """
    Attributes:
     - numStatementsParsed
     - numStatementsSucceeded
     - numStatementsFailed
     - executionTime
    """


    def __init__(self, numStatementsParsed=None, numStatementsSucceeded=None, numStatementsFailed=None, executionTime=None,):
        self.numStatementsParsed = numStatementsParsed
        self.numStatementsSucceeded = numStatementsSucceeded
        self.numStatementsFailed = numStatementsFailed
        self.executionTime = executionTime

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I64:
                    self.numStatementsParsed = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I64:
                    self.numStatementsSucceeded = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I64:
                    self.numStatementsFailed = iprot.readI64()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.DOUBLE:
                    self.executionTime = iprot.readDouble()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('QueryExecuteResponse')
        if self.numStatementsParsed is not None:
            oprot.writeFieldBegin('numStatementsParsed', TType.I64, 1)
            oprot.writeI64(self.numStatementsParsed)
            oprot.writeFieldEnd()
        if self.numStatementsSucceeded is not None:
            oprot.writeFieldBegin('numStatementsSucceeded', TType.I64, 2)
            oprot.writeI64(self.numStatementsSucceeded)
            oprot.writeFieldEnd()
        if self.numStatementsFailed is not None:
            oprot.writeFieldBegin('numStatementsFailed', TType.I64, 3)
            oprot.writeI64(self.numStatementsFailed)
            oprot.writeFieldEnd()
        if self.executionTime is not None:
            oprot.writeFieldBegin('executionTime', TType.DOUBLE, 4)
            oprot.writeDouble(self.executionTime)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.numStatementsParsed is None:
            raise TProtocolException(message='Required field numStatementsParsed is unset!')
        if self.numStatementsSucceeded is None:
            raise TProtocolException(message='Required field numStatementsSucceeded is unset!')
        if self.numStatementsFailed is None:
            raise TProtocolException(message='Required field numStatementsFailed is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(QueryExecuteResponse)
QueryExecuteResponse.thrift_spec = (
    None,  # 0
    (1, TType.I64, 'numStatementsParsed', None, None, ),  # 1
    (2, TType.I64, 'numStatementsSucceeded', None, None, ),  # 2
    (3, TType.I64, 'numStatementsFailed', None, None, ),  # 3
    (4, TType.DOUBLE, 'executionTime', None, None, ),  # 4
)
fix_spec(all_structs)
del all_structs



from setuptools import find_namespace_packages, setup

version = "4.2.0"
long_description = open("README.md").read()

setup(
    name="pyrdpdb",
    version=version,
    license="MIT",
    url="",
    author="Robert Li",
    author_email="robert.li@boraydata.com",
    description="Pure Python RapidsDB Driver",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(where=".", include=["pyrdpdb*", "pyrdpdb.*"]),
    package_dir={"pyrdpdb": "pyrdpdb"},
    entry_points={
        "sqlalchemy.dialects": [
            "rapidsdb.pyrdp = pyrdpdb.pyrdp.sa.pyrdp:RDPDialect_pyrdp",
            "rapidsdb.asyncrdp = pyrdpdb.aiordp.sa.asyncrapids:RDPDialect_asyncrdp",
        ]
    },
    install_requires=[
        "thrift==0.11.0",
        "SQLAlchemy>=2.0, <3.0",
    ],
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Topic :: Database",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class MessageTypes(object):
    ProtocolCompatibleRequest = 1
    ProtocolCompatibleResponse = 2
    ProtocolIncompatibleResponse = 4

    _VALUES_TO_NAMES = {
        1: "ProtocolCompatibleRequest",
        2: "ProtocolCompatibleResponse",
        4: "ProtocolIncompatibleResponse",
    }

    _NAMES_TO_VALUES = {
        "ProtocolCompatibleRequest": 1,
        "ProtocolCompatibleResponse": 2,
        "ProtocolIncompatibleResponse": 4,
    }
fix_spec(all_structs)
del all_structs



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class MessageTypes(object):
    ProtocolCompatibleRequest = 1
    ProtocolCompatibleResponse = 2
    ProtocolIncompatibleResponse = 4
    ErrorResponse = 6
    WarningResponse = 8
    ShutdownNotification = 10
    DisconnectNotification = 12
    DisconnectRequest = 13
    ClearTextTunnelRequest = 21
    TunnelSupportedResponse = 22
    AuthenticationRequest = 23
    AuthenticatePasswordRequest = 24
    AuthenticatePasswordResponse = 25
    AuthenticationOkResponse = 26
    SessionChangeRequest = 27
    SessionChangeResponse = 28
    QueryExecuteRequest = 29
    QueryReceiptResponse = 30
    RowMetadataResponse = 32
    RowDataResponse = 34
    StatementResponse = 36
    QueryExecuteResponse = 38
    QueryCancellationRequest = 39
    QueryCancellationResponse = 40
    QueryProgressRequest = 41
    QueryProgressResponse = 42
    SessionMonitoringRequest = 43
    SessionProtocolVersionResponse = 44
    SessionDataResponse = 46
    SessionDroppedMessagesResponse = 48
    SessionMonitoringResponse = 50
    ParseStatementRequest = 51
    ParseStatementResponse = 52
    StatementBindAndExecuteRequest = 53
    ClosePreparedStatementRequest = 55
    ClosePreparedStatementResponse = 56

    _VALUES_TO_NAMES = {
        1: "ProtocolCompatibleRequest",
        2: "ProtocolCompatibleResponse",
        4: "ProtocolIncompatibleResponse",
        6: "ErrorResponse",
        8: "WarningResponse",
        10: "ShutdownNotification",
        12: "DisconnectNotification",
        13: "DisconnectRequest",
        21: "ClearTextTunnelRequest",
        22: "TunnelSupportedResponse",
        23: "AuthenticationRequest",
        24: "AuthenticatePasswordRequest",
        25: "AuthenticatePasswordResponse",
        26: "AuthenticationOkResponse",
        27: "SessionChangeRequest",
        28: "SessionChangeResponse",
        29: "QueryExecuteRequest",
        30: "QueryReceiptResponse",
        32: "RowMetadataResponse",
        34: "RowDataResponse",
        36: "StatementResponse",
        38: "QueryExecuteResponse",
        39: "QueryCancellationRequest",
        40: "QueryCancellationResponse",
        41: "QueryProgressRequest",
        42: "QueryProgressResponse",
        43: "SessionMonitoringRequest",
        44: "SessionProtocolVersionResponse",
        46: "SessionDataResponse",
        48: "SessionDroppedMessagesResponse",
        50: "SessionMonitoringResponse",
        51: "ParseStatementRequest",
        52: "ParseStatementResponse",
        53: "StatementBindAndExecuteRequest",
        55: "ClosePreparedStatementRequest",
        56: "ClosePreparedStatementResponse",
    }

    _NAMES_TO_VALUES = {
        "ProtocolCompatibleRequest": 1,
        "ProtocolCompatibleResponse": 2,
        "ProtocolIncompatibleResponse": 4,
        "ErrorResponse": 6,
        "WarningResponse": 8,
        "ShutdownNotification": 10,
        "DisconnectNotification": 12,
        "DisconnectRequest": 13,
        "ClearTextTunnelRequest": 21,
        "TunnelSupportedResponse": 22,
        "AuthenticationRequest": 23,
        "AuthenticatePasswordRequest": 24,
        "AuthenticatePasswordResponse": 25,
        "AuthenticationOkResponse": 26,
        "SessionChangeRequest": 27,
        "SessionChangeResponse": 28,
        "QueryExecuteRequest": 29,
        "QueryReceiptResponse": 30,
        "RowMetadataResponse": 32,
        "RowDataResponse": 34,
        "StatementResponse": 36,
        "QueryExecuteResponse": 38,
        "QueryCancellationRequest": 39,
        "QueryCancellationResponse": 40,
        "QueryProgressRequest": 41,
        "QueryProgressResponse": 42,
        "SessionMonitoringRequest": 43,
        "SessionProtocolVersionResponse": 44,
        "SessionDataResponse": 46,
        "SessionDroppedMessagesResponse": 48,
        "SessionMonitoringResponse": 50,
        "ParseStatementRequest": 51,
        "ParseStatementResponse": 52,
        "StatementBindAndExecuteRequest": 53,
        "ClosePreparedStatementRequest": 55,
        "ClosePreparedStatementResponse": 56,
    }
fix_spec(all_structs)
del all_structs
